/* This software was developed by Bruce Hendrickson and Robert Leland   *
 * at Sandia National Laboratories under US Department of Energy        *
 * contract DE-AC04-76DP00789 and is copyrighted by Sandia Corporation. */

#include	<stdio.h>
#include	"structs.h"
#include	"defs.h"

/* Refine a vertex separator by finding a maximum bipartite matching. */

static void bpm_improve1();

void      bpm_improve(graph, sets, goal, max_dev, bndy_list, weights)
struct vtx_data **graph;	/* list of graph info for each vertex */
short    *sets;			/* local partitioning of vtxs */
double   *goal;			/* desired set sizes */
int       max_dev;		/* largest deviation from balance allowed */
int     **bndy_list;		/* list of vertices on boundary (0 ends) */
double   *weights;		/* vertex weights in each set */
{
    double    ratio;		/* fraction of non-separator vertices */
    double    deltaplus;	/* amount set is too big */
    double    deltaminus;	/* amount set is too small */
    double    imbalance;	/* current amount of imbalance */
    int       set_big;		/* side of graph I'm matching against */
    int       set_small;	/* side of graph I'm not matching against */
    int       sep_size;		/* separator size */
    int       old_sep_size;	/* previous separator size */
    int       change;		/* does separator get improved? */
    double    fabs();

    sep_size = 0;
    while ((*bndy_list)[sep_size] != 0) {
	sep_size++;
    }

/*
    printf("Before first matching, sep_size = %d,  Sizes = %g/%g\n",
	   sep_size, weights[0], weights[1]);
*/

    ratio = (weights[0] + weights[1]) / (goal[0] + goal[1]);
    deltaplus = fabs(weights[0] - goal[0] * ratio);
    deltaminus = fabs(weights[1] - goal[1] * ratio);
    imbalance = deltaplus + deltaminus;


    change = TRUE;
    while (change) {
	/* First match towards the larger side, then the smaller. */
	if (goal[0] - weights[0] >= goal[1] - weights[1]) {
	    set_big = 1;
	    set_small = 0;
	}
	else {
	    set_big = 0;
	    set_small = 1;
	}

	old_sep_size = sep_size;
	bpm_improve1(graph, sets, bndy_list, weights, set_big, set_small,
		     goal, max_dev, &imbalance, &sep_size);

	change = (sep_size != old_sep_size);

/*
	printf("After big matching, sep_size = %d,  Sizes = %g/%g\n",
	       sep_size, weights[0], weights[1]);
*/

	if (!change) {
	    /* If balanced, try the other direction. */
	    if (imbalance < max_dev) {
		bpm_improve1(graph, sets, bndy_list, weights, set_small, set_big,
			     goal, max_dev, &imbalance, &sep_size);

		change = (sep_size != old_sep_size);
/*
		printf("After small matching, sep_size = %d,  Sizes = %g/%g\n",
		       sep_size, weights[0], weights[1]);
*/
	    }
	}
    }
/*
    printf("After all matchings, sep_size = %d,  Sizes = %g/%g\n\n",
       sep_size, weights[0], weights[1]);
*/
}


static void bpm_improve1(graph, sets, pbndy_list, weights, set_match, set_other,
		       goal, max_dev, pimbalance, sep_size)
struct vtx_data **graph;	/* list of graph info for each vertex */
short    *sets;			/* local partitioning of vtxs */
int     **pbndy_list;		/* list of vertices on boundary (0 ends) */
double   *weights;		/* vertex weights in each set */
int       set_match;		/* side of graph I'm matching against */
int       set_other;		/* side of graph I'm not matching against */
double   *goal;			/* desired set sizes */
int       max_dev;		/* largest deviation from balance allowed */
double   *pimbalance;		/* imbalance of current partition */
int      *sep_size;		/* separator size */
{
    double    new_weights[2];	/* weights associated with new separator */
    double    ratio;		/* fraction of non-separator vertices */
    double    deltaplus;	/* amount set is too big */
    double    deltaminus;	/* amount set is too small */
    double    new_imbalance;	/* imbalance of new partition */
    int      *pointers;		/* start/stop indices into adjacencies */
    int      *indices;		/* adjacencies for each bipartite vertex */
    int      *loc2glob;		/* mapping from bp graph to original */
    int      *new_bndy_list;	/* new list of boundary vertices */
    int       old_sep_size;	/* previous separator size */
    int       vtx;		/* vertex in graph */
    int       nleft, nright;	/* # vtxs in two sides on bp graph */
    int       i, j;		/* loop counter */
    void      make_bpgraph(), bpcover();
    double   *smalloc(), fabs();
    int       sfree();

    make_bpgraph(graph, sets, *pbndy_list, *sep_size, set_match,
		 &pointers, &indices, &loc2glob, &nleft, &nright);

    new_bndy_list = (int *) smalloc((unsigned) (*sep_size + 1) * sizeof(int));
    new_bndy_list[0] = nleft + nright;
    old_sep_size = *sep_size;
    bpcover(nleft, nright, pointers, indices, sep_size, new_bndy_list);

    /* Update weights. */
    new_weights[0] = weights[0];
    new_weights[1] = weights[1];
    for (j = 0; j < new_bndy_list[0]; j++) {
	/* First handle nodes numbered  less than separator nodes. */
	vtx = loc2glob[j];
	if (sets[vtx] == 2) {
	    new_weights[set_other] += graph[vtx]->vwgt;
	}
    }
    for (i = 0; i < *sep_size; i++) {
	vtx = loc2glob[new_bndy_list[i]];
	if (sets[vtx] == set_match) {
	    new_weights[set_match] -= graph[vtx]->vwgt;
	}
	if (i != 0) {
	    for (j = new_bndy_list[i - 1] + 1; j < new_bndy_list[i]; j++) {
		vtx = loc2glob[j];
		if (sets[vtx] == 2) {
		    new_weights[set_other] += graph[vtx]->vwgt;
		}
	    }
	}
    }
    if (*sep_size != 0) {
	for (j = new_bndy_list[*sep_size - 1] + 1; j < nleft + nright; j++) {
	    vtx = loc2glob[j];
	    if (sets[vtx] == 2) {
		new_weights[set_other] += graph[vtx]->vwgt;
	    }
	}
    }


    /* Check to see if new partition is acceptably balanced. */
    ratio = (new_weights[0] + new_weights[1]) / (goal[0] + goal[1]);
    deltaplus = fabs(new_weights[0] - goal[0] * ratio);
    deltaminus = fabs(new_weights[1] - goal[1] * ratio);
    new_imbalance = deltaplus + deltaminus;

    if (new_imbalance <= max_dev || new_imbalance <= *pimbalance) {
	/* Update set values. */
	for (j = 0; j < new_bndy_list[0]; j++) {
	    /* First handle nodes numbered  less than separator nodes. */
	    vtx = loc2glob[j];
	    if (sets[vtx] == 2) {
		sets[vtx] = (short) set_other;
	    }
	}
	for (i = 0; i < *sep_size; i++) {
	    vtx = loc2glob[new_bndy_list[i]];
	    if (sets[vtx] == set_match) {
		sets[vtx] = 2;
	    }
	    if (i != 0) {
		for (j = new_bndy_list[i - 1] + 1; j < new_bndy_list[i]; j++) {
		    vtx = loc2glob[j];
		    if (sets[vtx] == 2) {
			sets[vtx] = (short) set_other;
		    }
		}
	    }
	}
	if (*sep_size != 0) {
	    for (j = new_bndy_list[*sep_size - 1] + 1; j < nleft + nright; j++) {
		vtx = loc2glob[j];
		if (sets[vtx] == 2) {
		    sets[vtx] = (short) set_other;
		}
	    }
	}

	/* Restore bndy_list to global numbering. */
	for (i = 0; i < *sep_size; i++) {
	    new_bndy_list[i] = loc2glob[new_bndy_list[i]];
	}
	new_bndy_list[*sep_size] = 0;

	sfree((char *) *pbndy_list);

	*pbndy_list = new_bndy_list;
	*pimbalance = new_imbalance;

    }

    else {
	sfree((char *) new_bndy_list);
	*sep_size = old_sep_size;
    }

    sfree((char *) loc2glob);
    sfree((char *) indices);
    sfree((char *) pointers);
}
